@@ -38,18 +38,23 @@
 #include "attrs.h"
 #include "buffer.h"
 #include "compat.h"
+#include "constants.h"
 #include "debug.h"
+#include "hash.h"
 #include "pkcs11.h"
 #include "pkcs11x.h"
 
 #include <assert.h>
 #include <stdarg.h>
 #include <stdio.h>
+#include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
 
+#define ELEMS(x) (sizeof (x) / sizeof (x[0]))
+
 bool
-p11_attrs_is_empty (const CK_ATTRIBUTE *attrs)
+p11_attrs_terminator (const CK_ATTRIBUTE *attrs)
 {
 	return (attrs == NULL || attrs->type == CKA_INVALID);
 }
@@ -62,7 +67,7 @@ p11_attrs_count (const CK_ATTRIBUTE *attrs)
 	if (attrs == NULL)
 		return 0UL;
 
-	for (count = 0; !p11_attrs_is_empty (attrs); count++, attrs++);
+	for (count = 0; !p11_attrs_terminator (attrs); count++, attrs++);
 
 	return count;
 }
@@ -76,7 +81,7 @@ p11_attrs_free (void *attrs)
 	if (!attrs)
 		return;
 
-	for (i = 0; !p11_attrs_is_empty (ats + i); i++)
+	for (i = 0; !p11_attrs_terminator (ats + i); i++)
 		free (ats[i].pValue);
 	free (ats);
 }
@@ -84,7 +89,8 @@ p11_attrs_free (void *attrs)
 static CK_ATTRIBUTE *
 attrs_build (CK_ATTRIBUTE *attrs,
              CK_ULONG count_to_add,
-             bool copy,
+             bool take_values,
+             bool override,
              CK_ATTRIBUTE * (*generator) (void *),
              void *state)
 {
@@ -116,24 +122,34 @@ attrs_build (CK_ATTRIBUTE *attrs,
 		for (j = 0; attr == NULL && j < current; j++) {
 			if (attrs[j].type == add->type) {
 				attr = attrs + j;
-				free (attrs[j].pValue);
 				break;
 			}
 		}
 
+		/* The attribute doesn't exist */
 		if (attr == NULL) {
 			attr = attrs + at;
 			at++;
+
+		/* The attribute exists and we're not overriding */
+		} else if (!override) {
+			if (take_values)
+				free (add->pValue);
+			continue;
+
+		/* The attribute exitss, and we're overriding */
+		} else {
+			free (attr->pValue);
 		}
 
 		memcpy (attr, add, sizeof (CK_ATTRIBUTE));
-		if (copy)
+		if (!take_values)
 			attr->pValue = memdup (attr->pValue, attr->ulValueLen);
 	}
 
 	/* Mark this as the end */
 	(attrs + at)->type = CKA_INVALID;
-	assert (p11_attrs_is_empty (attrs + at));
+	assert (p11_attrs_terminator (attrs + at));
 	return attrs;
 }
 
@@ -158,7 +174,8 @@ p11_attrs_build (CK_ATTRIBUTE *attrs,
 	va_end (va);
 
 	va_start (va, attrs);
-	attrs = attrs_build (attrs, count, true, vararg_generator, &va);
+	attrs = attrs_build (attrs, count, false, true,
+	                     vararg_generator, &va);
 	va_end (va);
 
 	return attrs;
@@ -176,7 +193,8 @@ p11_attrs_buildn (CK_ATTRIBUTE *attrs,
                   const CK_ATTRIBUTE *add,
                   CK_ULONG count)
 {
-	return attrs_build (attrs, count, true, template_generator, &add);
+	return attrs_build (attrs, count, false, true,
+	                    template_generator, &add);
 }
 
 CK_ATTRIBUTE *
@@ -187,7 +205,34 @@ p11_attrs_take (CK_ATTRIBUTE *attrs,
 {
 	CK_ATTRIBUTE attr = { type, value, length };
 	CK_ATTRIBUTE *add = &attr;
-	return attrs_build (attrs, 1, false, template_generator, &add);
+	return attrs_build (attrs, 1, true, true,
+	                    template_generator, &add);
+}
+
+CK_ATTRIBUTE *
+p11_attrs_merge (CK_ATTRIBUTE *attrs,
+                 CK_ATTRIBUTE *merge,
+                 bool replace)
+{
+	CK_ATTRIBUTE *ptr;
+	CK_ULONG count;
+
+	if (attrs == NULL)
+		return merge;
+
+	ptr = merge;
+	count = p11_attrs_count (merge);
+
+	attrs = attrs_build (attrs, count, true, replace,
+	                     template_generator, &ptr);
+
+	/*
+	 * Since we're supposed to own the merge attributes,
+	 * free the container array.
+	 */
+	free (merge);
+
+	return attrs;
 }
 
 CK_ATTRIBUTE *
@@ -205,7 +250,7 @@ p11_attrs_find (CK_ATTRIBUTE *attrs,
 {
 	CK_ULONG i;
 
-	for (i = 0; !p11_attrs_is_empty (attrs + i); i++) {
+	for (i = 0; !p11_attrs_terminator (attrs + i); i++) {
 		if (attrs[i].type == type)
 			return attrs + i;
 	}
@@ -229,13 +274,13 @@ p11_attrs_findn (CK_ATTRIBUTE *attrs,
 }
 
 bool
-p11_attrs_find_bool (CK_ATTRIBUTE *attrs,
+p11_attrs_find_bool (const CK_ATTRIBUTE *attrs,
                      CK_ATTRIBUTE_TYPE type,
                      CK_BBOOL *value)
 {
 	CK_ULONG i;
 
-	for (i = 0; !p11_attrs_is_empty (attrs + i); i++) {
+	for (i = 0; !p11_attrs_terminator (attrs + i); i++) {
 		if (attrs[i].type == type &&
 		    attrs[i].ulValueLen == sizeof (CK_BBOOL) &&
 		    attrs[i].pValue != NULL) {
@@ -248,7 +293,7 @@ p11_attrs_find_bool (CK_ATTRIBUTE *attrs,
 }
 
 bool
-p11_attrs_findn_bool (CK_ATTRIBUTE *attrs,
+p11_attrs_findn_bool (const CK_ATTRIBUTE *attrs,
                       CK_ULONG count,
                       CK_ATTRIBUTE_TYPE type,
                       CK_BBOOL *value)
@@ -268,13 +313,13 @@ p11_attrs_findn_bool (CK_ATTRIBUTE *attrs,
 }
 
 bool
-p11_attrs_find_ulong (CK_ATTRIBUTE *attrs,
+p11_attrs_find_ulong (const CK_ATTRIBUTE *attrs,
                       CK_ATTRIBUTE_TYPE type,
                       CK_ULONG *value)
 {
 	CK_ULONG i;
 
-	for (i = 0; !p11_attrs_is_empty (attrs + i); i++) {
+	for (i = 0; !p11_attrs_terminator (attrs + i); i++) {
 		if (attrs[i].type == type &&
 		    attrs[i].ulValueLen == sizeof (CK_ULONG) &&
 		    attrs[i].pValue != NULL) {
@@ -287,7 +332,7 @@ p11_attrs_find_ulong (CK_ATTRIBUTE *attrs,
 }
 
 bool
-p11_attrs_findn_ulong (CK_ATTRIBUTE *attrs,
+p11_attrs_findn_ulong (const CK_ATTRIBUTE *attrs,
                        CK_ULONG count,
                        CK_ATTRIBUTE_TYPE type,
                        CK_ULONG *value)
@@ -306,30 +351,37 @@ p11_attrs_findn_ulong (CK_ATTRIBUTE *attrs,
 	return false;
 }
 
-CK_ATTRIBUTE *
-p11_attrs_find_valid (CK_ATTRIBUTE *attrs,
-                      CK_ATTRIBUTE_TYPE type)
+void *
+p11_attrs_find_value (CK_ATTRIBUTE *attrs,
+                      CK_ATTRIBUTE_TYPE type,
+                      size_t *length)
 {
 	CK_ULONG i;
 
-	for (i = 0; !p11_attrs_is_empty (attrs + i); i++) {
+	for (i = 0; !p11_attrs_terminator (attrs + i); i++) {
 		if (attrs[i].type == type &&
-		    attrs[i].ulValueLen != (CK_ULONG)-1)
-			return attrs + i;
+		    attrs[i].ulValueLen != 0 &&
+		    attrs[i].ulValueLen != (CK_ULONG)-1 &&
+		    attrs[i].pValue != NULL) {
+			if (length)
+				*length = attrs[i].ulValueLen;
+			return attrs[i].pValue;
+		}
 	}
 
 	return NULL;
 }
 
 CK_ATTRIBUTE *
-p11_attrs_findn_valid (CK_ATTRIBUTE *attrs,
-                       CK_ULONG count,
-                       CK_ATTRIBUTE_TYPE type)
+p11_attrs_find_valid (CK_ATTRIBUTE *attrs,
+                      CK_ATTRIBUTE_TYPE type)
 {
 	CK_ULONG i;
 
-	for (i = 0; i < count; i++) {
+	for (i = 0; !p11_attrs_terminator (attrs + i); i++) {
 		if (attrs[i].type == type &&
+		    attrs[i].pValue != NULL &&
+		    attrs[i].ulValueLen != 0 &&
 		    attrs[i].ulValueLen != (CK_ULONG)-1)
 			return attrs + i;
 	}
@@ -361,13 +413,35 @@ p11_attrs_remove (CK_ATTRIBUTE *attrs,
 	return true;
 }
 
+void
+p11_attrs_purge (CK_ATTRIBUTE *attrs)
+{
+	int in, out;
+
+	for (in = 0, out = 0; !p11_attrs_terminator (attrs + in); in++) {
+		if (attrs[in].ulValueLen == (CK_ULONG)-1) {
+			free (attrs[in].pValue);
+			attrs[in].pValue = NULL;
+			attrs[in].ulValueLen = 0;
+		} else {
+			if (in != out)
+				memcpy (attrs + out, attrs + in, sizeof (CK_ATTRIBUTE));
+			out++;
+		}
+	}
+
+	attrs[out].type = CKA_INVALID;
+	assert (p11_attrs_terminator (attrs + out));
+
+}
+
 bool
 p11_attrs_match (const CK_ATTRIBUTE *attrs,
                  const CK_ATTRIBUTE *match)
 {
 	CK_ATTRIBUTE *attr;
 
-	for (; !p11_attrs_is_empty (match); match++) {
+	for (; !p11_attrs_terminator (match); match++) {
 		attr = p11_attrs_find ((CK_ATTRIBUTE *)attrs, match->type);
 		if (!attr)
 			return false;
@@ -429,11 +503,12 @@ unsigned int
 p11_attr_hash (const void *data)
 {
 	const CK_ATTRIBUTE *attr = data;
-	unsigned int hash = (unsigned int)attr->type;
-	const char *p, *end;
+	uint32_t hash;
 
-	for (p = attr->pValue, end = p + attr->ulValueLen ; p != NULL && p != end; p++)
-		hash = (hash << 5) - hash + *p;
+	p11_hash_murmur2 (&hash,
+	                  &attr->type, sizeof (attr->type),
+	                  attr->pValue, (size_t)attr->ulValueLen,
+	                  NULL);
 
 	return hash;
 }
@@ -496,7 +571,8 @@ attribute_is_trust_value (const CK_ATTRIBUTE *attr)
 }
 
 static bool
-attribute_is_sensitive (const CK_ATTRIBUTE *attr)
+attribute_is_sensitive (const CK_ATTRIBUTE *attr,
+                        CK_OBJECT_CLASS klass)
 {
 	/*
 	 * Don't print any just attribute, since they may contain
@@ -612,6 +688,9 @@ attribute_is_sensitive (const CK_ATTRIBUTE *attr)
 	X (CKA_TRUST_STEP_UP_APPROVED)
 	X (CKA_CERT_SHA1_HASH)
 	X (CKA_CERT_MD5_HASH)
+	case CKA_VALUE:
+		return (klass != CKO_CERTIFICATE &&
+			klass != CKO_X_CERTIFICATE_EXTENSION);
 	#undef X
 	}
 
@@ -622,29 +701,7 @@ static void
 format_class (p11_buffer *buffer,
               CK_OBJECT_CLASS klass)
 {
-	const char *string = NULL;
-
-	switch (klass) {
-	#define X(x) case x: string = #x; break;
-	X (CKO_DATA)
-	X (CKO_CERTIFICATE)
-	X (CKO_PUBLIC_KEY)
-	X (CKO_PRIVATE_KEY)
-	X (CKO_SECRET_KEY)
-	X (CKO_HW_FEATURE)
-	X (CKO_DOMAIN_PARAMETERS)
-	X (CKO_MECHANISM)
-	X (CKO_X_TRUST_ASSERTION)
-	X (CKO_X_CERTIFICATE_EXTENSION)
-	X (CKO_NSS_CRL)
-	X (CKO_NSS_SMIME)
-	X (CKO_NSS_TRUST)
-	X (CKO_NSS_BUILTIN_ROOT_LIST)
-	X (CKO_NSS_NEWSLOT)
-	X (CKO_NSS_DELSLOT)
-	#undef X
-	}
-
+	const char *string = p11_constant_name (p11_constant_classes, klass);
 	if (string != NULL)
 		p11_buffer_add (buffer, string, -1);
 	else
@@ -655,16 +712,7 @@ static void
 format_assertion_type (p11_buffer *buffer,
                        CK_X_ASSERTION_TYPE type)
 {
-	const char *string = NULL;
-
-	switch (type) {
-	#define X(x) case x: string = #x; break;
-	X (CKT_X_DISTRUSTED_CERTIFICATE)
-	X (CKT_X_PINNED_CERTIFICATE)
-	X (CKT_X_ANCHORED_CERTIFICATE)
-	#undef X
-	}
-
+	const char *string = p11_constant_name (p11_constant_asserts, type);
 	if (string != NULL)
 		p11_buffer_add (buffer, string, -1);
 	else
@@ -675,39 +723,7 @@ static void
 format_key_type (p11_buffer *buffer,
                  CK_KEY_TYPE type)
 {
-	const char *string = NULL;
-
-	switch (type) {
-	#define X(x) case x: string = #x; break;
-	X (CKK_RSA)
-	X (CKK_DSA)
-	X (CKK_DH)
-	/* X (CKK_ECDSA) */
-	X (CKK_EC)
-	X (CKK_X9_42_DH)
-	X (CKK_KEA)
-	X (CKK_GENERIC_SECRET)
-	X (CKK_RC2)
-	X (CKK_RC4)
-	X (CKK_DES)
-	X (CKK_DES2)
-	X (CKK_DES3)
-	X (CKK_CAST)
-	X (CKK_CAST3)
-	X (CKK_CAST128)
-	X (CKK_RC5)
-	X (CKK_IDEA)
-	X (CKK_SKIPJACK)
-	X (CKK_BATON)
-	X (CKK_JUNIPER)
-	X (CKK_CDMF)
-	X (CKK_AES)
-	X (CKK_BLOWFISH)
-	X (CKK_TWOFISH)
-	X (CKK_NSS_PKCS8)
-	#undef X
-	}
-
+	const char *string = p11_constant_name (p11_constant_keys, type);
 	if (string != NULL)
 		p11_buffer_add (buffer, string, -1);
 	else
@@ -718,15 +734,7 @@ static void
 format_certificate_type (p11_buffer *buffer,
                          CK_CERTIFICATE_TYPE type)
 {
-	const char *string = NULL;
-
-	switch (type) {
-	#define X(x) case x: string = #x; break;
-	X (CKC_X_509)
-	X (CKC_X_509_ATTR_CERT)
-	X (CKC_WTLS)
-	}
-
+	const char *string = p11_constant_name (p11_constant_certs, type);
 	if (string != NULL)
 		p11_buffer_add (buffer, string, -1);
 	else
@@ -737,17 +745,7 @@ static void
 format_trust_value (p11_buffer *buffer,
                     CK_TRUST trust)
 {
-	const char *string = NULL;
-
-	switch (trust) {
-	#define X(x) case x: string = #x; break;
-	X (CKT_NSS_TRUSTED)
-	X (CKT_NSS_TRUSTED_DELEGATOR)
-	X (CKT_NSS_NOT_TRUSTED)
-	X (CKT_NSS_MUST_VERIFY_TRUST)
-	X (CKT_NSS_TRUST_UNKNOWN)
-	}
-
+	const char *string = p11_constant_name (p11_constant_trusts, trust);
 	if (string != NULL)
 		p11_buffer_add (buffer, string, -1);
 	else
@@ -758,23 +756,7 @@ static void
 format_certificate_category (p11_buffer *buffer,
                              CK_ULONG category)
 {
-	const char *string = NULL;
-
-	switch (category) {
-	case 0:
-		string = "unspecified";
-		break;
-	case 1:
-		string = "token-user";
-		break;
-	case 2:
-		string = "authority";
-		break;
-	case 3:
-		string = "other-entry";
-		break;
-	}
-
+	const char *string = p11_constant_name (p11_constant_categories, category);
 	if (string != NULL)
 		buffer_append_printf (buffer, "%lu (%s)", category, string);
 	else
@@ -785,135 +767,7 @@ static void
 format_attribute_type (p11_buffer *buffer,
                        CK_ULONG type)
 {
-	const char *string = NULL;
-
-	switch (type) {
-	#define X(x) case x: string = #x; break;
-	X (CKA_CLASS)
-	X (CKA_TOKEN)
-	X (CKA_PRIVATE)
-	X (CKA_LABEL)
-	X (CKA_APPLICATION)
-	X (CKA_VALUE)
-	X (CKA_OBJECT_ID)
-	X (CKA_CERTIFICATE_TYPE)
-	X (CKA_ISSUER)
-	X (CKA_SERIAL_NUMBER)
-	X (CKA_AC_ISSUER)
-	X (CKA_OWNER)
-	X (CKA_ATTR_TYPES)
-	X (CKA_TRUSTED)
-	X (CKA_CERTIFICATE_CATEGORY)
-	X (CKA_JAVA_MIDP_SECURITY_DOMAIN)
-	X (CKA_URL)
-	X (CKA_HASH_OF_SUBJECT_PUBLIC_KEY)
-	X (CKA_HASH_OF_ISSUER_PUBLIC_KEY)
-	X (CKA_CHECK_VALUE)
-	X (CKA_KEY_TYPE)
-	X (CKA_SUBJECT)
-	X (CKA_ID)
-	X (CKA_SENSITIVE)
-	X (CKA_ENCRYPT)
-	X (CKA_DECRYPT)
-	X (CKA_WRAP)
-	X (CKA_UNWRAP)
-	X (CKA_SIGN)
-	X (CKA_SIGN_RECOVER)
-	X (CKA_VERIFY)
-	X (CKA_VERIFY_RECOVER)
-	X (CKA_DERIVE)
-	X (CKA_START_DATE)
-	X (CKA_END_DATE)
-	X (CKA_MODULUS)
-	X (CKA_MODULUS_BITS)
-	X (CKA_PUBLIC_EXPONENT)
-	X (CKA_PRIVATE_EXPONENT)
-	X (CKA_PRIME_1)
-	X (CKA_PRIME_2)
-	X (CKA_EXPONENT_1)
-	X (CKA_EXPONENT_2)
-	X (CKA_COEFFICIENT)
-	X (CKA_PRIME)
-	X (CKA_SUBPRIME)
-	X (CKA_BASE)
-	X (CKA_PRIME_BITS)
-	/* X (CKA_SUBPRIME_BITS) */
-	X (CKA_SUB_PRIME_BITS)
-	X (CKA_VALUE_BITS)
-	X (CKA_VALUE_LEN)
-	X (CKA_EXTRACTABLE)
-	X (CKA_LOCAL)
-	X (CKA_NEVER_EXTRACTABLE)
-	X (CKA_ALWAYS_SENSITIVE)
-	X (CKA_KEY_GEN_MECHANISM)
-	X (CKA_MODIFIABLE)
-	X (CKA_ECDSA_PARAMS)
-	/* X (CKA_EC_PARAMS) */
-	X (CKA_EC_POINT)
-	X (CKA_SECONDARY_AUTH)
-	X (CKA_AUTH_PIN_FLAGS)
-	X (CKA_ALWAYS_AUTHENTICATE)
-	X (CKA_WRAP_WITH_TRUSTED)
-	X (CKA_WRAP_TEMPLATE)
-	X (CKA_UNWRAP_TEMPLATE)
-	X (CKA_HW_FEATURE_TYPE)
-	X (CKA_RESET_ON_INIT)
-	X (CKA_HAS_RESET)
-	X (CKA_PIXEL_X)
-	X (CKA_PIXEL_Y)
-	X (CKA_RESOLUTION)
-	X (CKA_CHAR_ROWS)
-	X (CKA_CHAR_COLUMNS)
-	X (CKA_COLOR)
-	X (CKA_BITS_PER_PIXEL)
-	X (CKA_CHAR_SETS)
-	X (CKA_ENCODING_METHODS)
-	X (CKA_MIME_TYPES)
-	X (CKA_MECHANISM_TYPE)
-	X (CKA_REQUIRED_CMS_ATTRIBUTES)
-	X (CKA_DEFAULT_CMS_ATTRIBUTES)
-	X (CKA_SUPPORTED_CMS_ATTRIBUTES)
-	X (CKA_ALLOWED_MECHANISMS)
-	X (CKA_X_ASSERTION_TYPE)
-	X (CKA_X_CERTIFICATE_VALUE)
-	X (CKA_X_PURPOSE)
-	X (CKA_X_PEER)
-	X (CKA_X_DISTRUSTED)
-	X (CKA_X_CRITICAL)
-	X (CKA_NSS_URL)
-	X (CKA_NSS_EMAIL)
-	X (CKA_NSS_SMIME_INFO)
-	X (CKA_NSS_SMIME_TIMESTAMP)
-	X (CKA_NSS_PKCS8_SALT)
-	X (CKA_NSS_PASSWORD_CHECK)
-	X (CKA_NSS_EXPIRES)
-	X (CKA_NSS_KRL)
-	X (CKA_NSS_PQG_COUNTER)
-	X (CKA_NSS_PQG_SEED)
-	X (CKA_NSS_PQG_H)
-	X (CKA_NSS_PQG_SEED_BITS)
-	X (CKA_NSS_MODULE_SPEC)
-	X (CKA_TRUST_DIGITAL_SIGNATURE)
-	X (CKA_TRUST_NON_REPUDIATION)
-	X (CKA_TRUST_KEY_ENCIPHERMENT)
-	X (CKA_TRUST_DATA_ENCIPHERMENT)
-	X (CKA_TRUST_KEY_AGREEMENT)
-	X (CKA_TRUST_KEY_CERT_SIGN)
-	X (CKA_TRUST_CRL_SIGN)
-	X (CKA_TRUST_SERVER_AUTH)
-	X (CKA_TRUST_CLIENT_AUTH)
-	X (CKA_TRUST_CODE_SIGNING)
-	X (CKA_TRUST_EMAIL_PROTECTION)
-	X (CKA_TRUST_IPSEC_END_SYSTEM)
-	X (CKA_TRUST_IPSEC_TUNNEL)
-	X (CKA_TRUST_IPSEC_USER)
-	X (CKA_TRUST_TIME_STAMPING)
-	X (CKA_TRUST_STEP_UP_APPROVED)
-	X (CKA_CERT_SHA1_HASH)
-	X (CKA_CERT_MD5_HASH)
-	#undef X
-	}
-
+	const char *string = p11_constant_name (p11_constant_types, type);
 	if (string != NULL)
 		p11_buffer_add (buffer, string, -1);
 	else
@@ -956,7 +810,8 @@ format_some_bytes (p11_buffer *buffer,
 
 static void
 format_attribute (p11_buffer *buffer,
-                  const CK_ATTRIBUTE *attr)
+                  const CK_ATTRIBUTE *attr,
+                  CK_OBJECT_CLASS klass)
 {
 	p11_buffer_add (buffer, "{ ", -1);
 	format_attribute_type (buffer, attr->type);
@@ -975,7 +830,7 @@ format_attribute (p11_buffer *buffer,
 		format_key_type (buffer, *((CK_KEY_TYPE *)attr->pValue));
 	} else if (attribute_is_trust_value (attr)) {
 		format_trust_value (buffer, *((CK_TRUST *)attr->pValue));
-	} else if (attribute_is_sensitive (attr)) {
+	} else if (attribute_is_sensitive (attr, klass)) {
 		buffer_append_printf (buffer, "(%lu) NOT-PRINTED", attr->ulValueLen);
 	} else {
 		buffer_append_printf (buffer, "(%lu) ", attr->ulValueLen);
@@ -986,12 +841,19 @@ format_attribute (p11_buffer *buffer,
 
 static void
 format_attributes (p11_buffer *buffer,
-                   const CK_ATTRIBUTE *attrs)
+                   const CK_ATTRIBUTE *attrs,
+                   int count)
 {
 	CK_BBOOL first = CK_TRUE;
-	int count, i;
+	CK_OBJECT_CLASS klass;
+	int i;
+
+	if (count < 0)
+		count = p11_attrs_count (attrs);
+
+	if (!p11_attrs_findn_ulong (attrs, CKA_CLASS, count, &klass))
+		klass = CKA_INVALID;
 
-	count = p11_attrs_count (attrs);
 	buffer_append_printf (buffer, "(%d) [", count);
 	for (i = 0; i < count; i++) {
 		if (first)
@@ -999,27 +861,29 @@ format_attributes (p11_buffer *buffer,
 		else
 			p11_buffer_add (buffer, ", ", 2);
 		first = CK_FALSE;
-		format_attribute (buffer, attrs + i);
+		format_attribute (buffer, attrs + i, klass);
 	}
 	p11_buffer_add (buffer, " ]", -1);
 }
 
 char *
-p11_attrs_to_string (const CK_ATTRIBUTE *attrs)
+p11_attrs_to_string (const CK_ATTRIBUTE *attrs,
+                     int count)
 {
 	p11_buffer buffer;
 	if (!p11_buffer_init_null (&buffer, 128))
 		return_val_if_reached (NULL);
-	format_attributes (&buffer, attrs);
+	format_attributes (&buffer, attrs, count);
 	return p11_buffer_steal (&buffer, NULL);
 }
 
 char *
-p11_attr_to_string (const CK_ATTRIBUTE *attr)
+p11_attr_to_string (const CK_ATTRIBUTE *attr,
+                    CK_OBJECT_CLASS klass)
 {
 	p11_buffer buffer;
 	if (!p11_buffer_init_null (&buffer, 32))
 		return_val_if_reached (NULL);
-	format_attribute (&buffer, attr);
+	format_attribute (&buffer, attr, klass);
 	return p11_buffer_steal (&buffer, NULL);
 }
