@@ -148,32 +148,40 @@ getprogname (void)
 
 #endif /* HAVE_GETPROGNAME */
 
-#ifndef HAVE_BASENAME
-
 char *
-basename (const char *name)
+p11_basename (const char *name)
 {
-	char *p;
 #ifdef OS_WIN32
-	char *p2;
+	static const char *delims = "/\\";
+#else
+	static const char *delims = "/";
 #endif
 
-	if (!name || name[0] == '\0')
-		return ".";
+	const char *end;
+	const char *beg;
 
-	p = strrchr (name, '/');
-#ifdef OS_WIN32
-	p2 = strrchr (name, '\\');
-	if (p2 > p)
-		p = p2;
-#endif
-	if (p != NULL)
-		return p + 1;
-	return (char *)name;
+	if (name == NULL)
+		return NULL;
+
+	/* Any trailing slashes */
+	end = name + strlen (name);
+	while (end != name) {
+		if (!strchr (delims, *(end - 1)))
+			break;
+		end--;
+	}
+
+	/* Find the last slash after those */
+	beg = end;
+	while (beg != name) {
+		if (strchr (delims, *(beg - 1)))
+			break;
+		beg--;
+	}
+
+	return strndup (beg, end - beg);
 }
 
-#endif /* HAVE_BASENAME */
-
 #ifdef OS_UNIX
 #include <sys/stat.h>
 #include <sys/mman.h>
@@ -424,27 +432,44 @@ memdup (const void *data,
 
 #endif /* HAVE_MEMDUP */
 
-#ifndef HAVE_STRNDUP
+/*
+ * WORKAROUND: So in lots of released builds of firefox a completely broken strndup()
+ * is present. It does not NULL terminate its string output. It is unconditionally
+ * defined, and overrides the libc strndup() function on platforms where it
+ * exists as a function. For this reason we (for now) unconditionally define
+ * strndup().
+ */
+
+#if 1 /* #ifndef HAVE_STRNDUP */
+
+/*
+ * HAVE_STRNDUP may be undefined if strndup() isn't working. So it may be
+ * present, and yet strndup may still be a defined header macro.
+ */
+#ifdef strndup
+#undef strndup
+#endif
+
+char *
+strndup (const char *data,
+         size_t length);
 
 char *
 strndup (const char *data,
          size_t length)
 {
 	char *ret;
-	size_t len;
 
-	len = strlen (data);
-	if (length > len)
-		length = len;
-
-	ret = memdup (data, length + 1);
-	if (ret != NULL)
+	ret = malloc (length + 1);
+	if (ret != NULL) {
+		strncpy (ret, data, length);
 		ret[length] = 0;
+	}
 
 	return ret;
 }
 
-#endif /* HAVE_STRDUP */
+#endif /* HAVE_STRNDUP */
 
 #ifndef HAVE_STRCONCAT
 
