@@ -34,6 +34,8 @@
 
 #include "config.h"
 
+#define P11_DEBUG_FLAG P11_DEBUG_TOOL
+
 #include "attrs.h"
 #include "debug.h"
 #include "oid.h"
@@ -115,19 +117,13 @@ extract_purposes (p11_extract_info *ex)
 	                     sizeof (P11_OID_EXTENDED_KEY_USAGE) };
 	const unsigned char *ext = NULL;
 	unsigned char *alloc = NULL;
-	CK_ATTRIBUTE *value;
 	CK_ATTRIBUTE *attrs;
 	size_t ext_len;
 
 	if (ex->stapled) {
 		attrs = p11_dict_get (ex->stapled, &oid);
-		if (attrs != NULL) {
-			value = p11_attrs_find (attrs, CKA_VALUE);
-			if (value) {
-				ext = value->pValue;
-				ext_len = value->ulValueLen;
-			}
-		}
+		if (attrs != NULL)
+			ext = p11_attrs_find_value (attrs, CKA_VALUE, &ext_len);
 	}
 
 	if (ext == NULL && ex->cert_asn) {
@@ -146,23 +142,95 @@ extract_purposes (p11_extract_info *ex)
 	return ex->purposes != NULL;
 }
 
+static bool
+should_collapse_certificate (p11_extract_info *ex,
+                             CK_ATTRIBUTE *value)
+{
+	CK_ATTRIBUTE *attrs;
+
+	if (!(ex->flags & P11_EXTRACT_COLLAPSE))
+		return false;
+
+	if (!ex->already_seen) {
+		ex->already_seen = p11_dict_new (p11_attr_hash, p11_attr_equal,
+		                                 NULL, p11_attrs_free);
+		return_val_if_fail (ex->already_seen != NULL, true);
+	}
+
+	if (p11_dict_get (ex->already_seen, value))
+		return true;
+
+	attrs = p11_attrs_build (NULL, value, NULL);
+	return_val_if_fail (attrs != NULL, true);
+
+	if (!p11_dict_set (ex->already_seen, attrs, attrs))
+		return_val_if_reached (true);
+
+	return false;
+}
+
+static bool
+check_trust_flags_match (p11_extract_info *ex)
+{
+	CK_BBOOL boolv;
+	int flags = 0;
+
+	/* If no extract trust flags, then just continue */
+	if (!(ex->flags & (P11_EXTRACT_ANCHORS | P11_EXTRACT_BLACKLIST)))
+		return true;
+
+	if (p11_attrs_find_bool (ex->attrs, CKA_TRUSTED, &boolv) && boolv)
+		flags |= P11_EXTRACT_ANCHORS;
+	if (p11_attrs_find_bool (ex->attrs, CKA_X_DISTRUSTED, &boolv) && boolv)
+		flags |= P11_EXTRACT_BLACKLIST;
+
+	/* Any of the flags can match */
+	if (flags & ex->flags)
+		return true;
+
+	return false;
+}
+
 static bool
 extract_certificate (P11KitIter *iter,
                      p11_extract_info *ex)
 {
 	char message[ASN1_MAX_ERROR_DESCRIPTION_SIZE];
 	CK_ATTRIBUTE *attr;
+
 	CK_ULONG type;
 
 	/* Don't even bother with not X.509 certificates */
 	if (!p11_attrs_find_ulong (ex->attrs, CKA_CERTIFICATE_TYPE, &type))
 		type = (CK_ULONG)-1;
-	if (type != CKC_X_509)
+	if (type != CKC_X_509) {
+		p11_debug ("skipping non X.509 certificate");
 		return false;
+	}
 
 	attr = p11_attrs_find_valid (ex->attrs, CKA_VALUE);
-	if (!attr || !attr->pValue)
+	if (!attr || !attr->pValue) {
+		p11_debug ("skipping certificate without a value");
 		return false;
+	}
+
+	/*
+	 * If collapsing and have already seen this certificate, and shouldn't
+	 * process it even again during this extract procedure.
+	 */
+	if (should_collapse_certificate (ex, attr)) {
+		p11_debug ("skipping certificate that has already been seen");
+		return false;
+	}
+
+	/*
+	 * We do these checks after collapsing, so that blacklisted certificates
+	 * mask out anchors even if we're not exporting blacklisted stuff.
+	 */
+	if (!check_trust_flags_match (ex)) {
+		p11_debug ("skipping certificate that doesn't match trust flags");
+		return false;
+	}
 
 	ex->cert_der = attr->pValue;
 	ex->cert_len = attr->ulValueLen;
@@ -207,14 +275,10 @@ extract_info (P11KitIter *iter,
 		return false;
 	}
 
-	attr = p11_attrs_find (ex->attrs, CKA_CLASS);
-
 	/* No class attribute, very strange, just skip */
-	if (!attr || !attr->pValue || attr->ulValueLen != sizeof (CK_OBJECT_CLASS))
+	if (!p11_attrs_find_ulong (ex->attrs, CKA_CLASS, &ex->klass))
 		return false;
 
-	ex->klass = *((CK_ULONG *)attr->pValue);
-
 	/* If a certificate then  */
 	if (ex->klass != CKO_CERTIFICATE) {
 		p11_message ("skipping non-certificate object");
@@ -224,7 +288,7 @@ extract_info (P11KitIter *iter,
 	if (!extract_certificate (iter, ex))
 		return false;
 
-	attr = p11_attrs_find (ex->attrs, CKA_ID);
+	attr = p11_attrs_find_valid (ex->attrs, CKA_ID);
 	if (attr) {
 		ex->stapled = load_stapled_extensions (p11_kit_iter_get_module (iter),
 		                                       p11_kit_iter_get_slot (iter),
@@ -308,6 +372,9 @@ p11_extract_info_cleanup (p11_extract_info *ex)
 	p11_dict_free (ex->limit_to_purposes);
 	ex->limit_to_purposes = NULL;
 
+	p11_dict_free (ex->already_seen);
+	ex->already_seen = NULL;
+
 	p11_dict_free (ex->asn1_defs);
 	ex->asn1_defs = NULL;
 }
@@ -336,7 +403,7 @@ extract_label (p11_extract_info *extract)
 	CK_ATTRIBUTE *attr;
 
 	/* Look for a label and just use that */
-	attr = p11_attrs_find (extract->attrs, CKA_LABEL);
+	attr = p11_attrs_find_valid (extract->attrs, CKA_LABEL);
 	if (attr && attr->pValue && attr->ulValueLen)
 		return strndup (attr->pValue, attr->ulValueLen);
 
@@ -366,3 +433,23 @@ p11_extract_info_filename (p11_extract_info *extract)
 
 	return label;
 }
+
+char *
+p11_extract_info_comment (p11_extract_info *ex,
+                          bool first)
+{
+	char *comment;
+	char *label;
+
+	if (!(ex->flags & P11_EXTRACT_COMMENT))
+		return NULL;
+
+	label = extract_label (ex);
+	if (!asprintf (&comment, "%s# %s\n",
+	               first ? "" : "\n",
+	               label ? label : ""))
+		return_val_if_reached (NULL);
+
+	free (label);
+	return comment;
+}
