@@ -37,11 +37,11 @@
 #include "asn1.h"
 #include "attrs.h"
 #include "buffer.h"
-#include "checksum.h"
 #include "compat.h"
 #include "debug.h"
 #include "dict.h"
 #include "extract.h"
+#include "hash.h"
 #include "library.h"
 #include "oid.h"
 #include "pem.h"
@@ -107,16 +107,16 @@ load_usage_ext (p11_extract_info *ex,
 {
 	CK_ATTRIBUTE attr = { CKA_OBJECT_ID, (void *)ext_oid,
 	                      p11_oid_length (ext_oid) };
-	CK_ATTRIBUTE *value;
+	void *value;
+	size_t length;
 
-	value = p11_attrs_find_valid (p11_dict_get (ex->stapled, &attr), CKA_VALUE);
+	value = p11_attrs_find_value (p11_dict_get (ex->stapled, &attr), CKA_VALUE, &length);
 	if (value == NULL) {
 		*oids = NULL;
 		return true;
 	}
 
-	*oids = p11_x509_parse_extended_key_usage (ex->asn1_defs, value->pValue,
-	                                           value->ulValueLen);
+	*oids = p11_x509_parse_extended_key_usage (ex->asn1_defs, value, length);
 	return_val_if_fail (*oids != NULL, false);
 
 	return true;
@@ -314,8 +314,10 @@ p11_extract_openssl_bundle (P11KitIter *iter,
 {
 	p11_save_file *file;
 	p11_buffer buf;
+	char *comment;
 	bool ret = true;
 	size_t length;
+	bool first;
 	CK_RV rv;
 	char *pem;
 
@@ -323,6 +325,7 @@ p11_extract_openssl_bundle (P11KitIter *iter,
 	if (!file)
 		return false;
 
+	first = true;
 	while ((rv = p11_kit_iter_next (iter)) == CKR_OK) {
 		p11_buffer_init (&buf, 1024);
 
@@ -330,8 +333,14 @@ p11_extract_openssl_bundle (P11KitIter *iter,
 			pem = p11_pem_write (buf.data, buf.len, "TRUSTED CERTIFICATE", &length);
 			return_val_if_fail (pem != NULL, false);
 
-			ret = p11_save_write (file, pem, length);
+			comment = p11_extract_info_comment (ex, first);
+			first = false;
+
+			ret = p11_save_write (file, comment, -1) &&
+			      p11_save_write (file, pem, length);
+
 			free (pem);
+			free (comment);
 		}
 
 		p11_buffer_uninit (&buf);
@@ -511,7 +520,7 @@ p11_openssl_canon_name_der (p11_dict *asn1_defs,
 static char *
 symlink_for_subject_hash (p11_extract_info *ex)
 {
-	unsigned char md[P11_CHECKSUM_SHA1_LENGTH];
+	unsigned char md[P11_HASH_SHA1_LEN];
 	p11_buffer der;
 	CK_ATTRIBUTE *subject;
 	unsigned long hash;
@@ -526,7 +535,7 @@ symlink_for_subject_hash (p11_extract_info *ex)
 	return_val_if_fail (der.data != NULL, NULL);
 
 	if (p11_openssl_canon_name_der (ex->asn1_defs, &der)) {
-		p11_checksum_sha1 (md, der.data, der.len, NULL);
+		p11_hash_sha1 (md, der.data, der.len, NULL);
 
 		hash = (
 		        ((unsigned long)md[0]       ) | ((unsigned long)md[1] << 8L) |
@@ -544,7 +553,7 @@ symlink_for_subject_hash (p11_extract_info *ex)
 static char *
 symlink_for_subject_old_hash (p11_extract_info *ex)
 {
-	unsigned char md[P11_CHECKSUM_MD5_LENGTH];
+	unsigned char md[P11_HASH_MD5_LEN];
 	CK_ATTRIBUTE *subject;
 	unsigned long hash;
 	char *linkname;
@@ -553,7 +562,7 @@ symlink_for_subject_old_hash (p11_extract_info *ex)
 	if (!subject)
 		return NULL;
 
-	p11_checksum_md5 (md, subject->pValue, subject->ulValueLen, NULL);
+	p11_hash_md5 (md, subject->pValue, subject->ulValueLen, NULL);
 
 	hash = (
 	         ((unsigned long)md[0]       ) | ((unsigned long)md[1] << 8L) |
@@ -593,9 +602,6 @@ p11_extract_openssl_directory (P11KitIter *iter,
 	p11_buffer_init (&buf, 0);
 
 	while ((rv = p11_kit_iter_next (iter)) == CKR_OK) {
-		pem = p11_pem_write (ex->cert_der, ex->cert_len, "CERTIFICATE", &length);
-		return_val_if_fail (pem != NULL, false);
-
 		if (!p11_buffer_reset (&buf, 1024))
 			return_val_if_reached (false);
 
