@@ -91,6 +91,49 @@ test_file_name_for_class (CuTest *tc)
 	p11_extract_info_cleanup (&ex);
 }
 
+static void
+test_comment_for_label (CuTest *tc)
+{
+	CK_ATTRIBUTE label = { CKA_LABEL, "The Label!", 10 };
+	p11_extract_info ex;
+	char *comment;
+
+	p11_extract_info_init (&ex);
+
+	ex.flags = P11_EXTRACT_COMMENT;
+	ex.attrs = p11_attrs_build (NULL, &label, NULL);
+
+	comment = p11_extract_info_comment (&ex, true);
+	CuAssertStrEquals (tc, "# The Label!\n", comment);
+	free (comment);
+
+	comment = p11_extract_info_comment (&ex, false);
+	CuAssertStrEquals (tc, "\n# The Label!\n", comment);
+	free (comment);
+
+	p11_extract_info_cleanup (&ex);
+}
+
+static void
+test_comment_not_enabled (CuTest *tc)
+{
+	CK_ATTRIBUTE label = { CKA_LABEL, "The Label!", 10 };
+	p11_extract_info ex;
+	char *comment;
+
+	p11_extract_info_init (&ex);
+
+	ex.attrs = p11_attrs_build (NULL, &label, NULL);
+
+	comment = p11_extract_info_comment (&ex, true);
+	CuAssertPtrEquals (tc, NULL, comment);
+
+	comment = p11_extract_info_comment (&ex, false);
+	CuAssertPtrEquals (tc, NULL, comment);
+
+	p11_extract_info_cleanup (&ex);
+}
+
 struct {
 	CK_FUNCTION_LIST module;
 	P11KitIter *iter;
@@ -128,17 +171,29 @@ teardown (CuTest *tc)
 static CK_OBJECT_CLASS certificate_class = CKO_CERTIFICATE;
 static CK_OBJECT_CLASS extension_class = CKO_X_CERTIFICATE_EXTENSION;
 static CK_CERTIFICATE_TYPE x509_type = CKC_X_509;
+static CK_BBOOL truev = CK_TRUE;
 
-static CK_ATTRIBUTE cacert3_authority_attrs[] = {
+static CK_ATTRIBUTE cacert3_trusted[] = {
 	{ CKA_VALUE, (void *)test_cacert3_ca_der, sizeof (test_cacert3_ca_der) },
 	{ CKA_CLASS, &certificate_class, sizeof (certificate_class) },
 	{ CKA_CERTIFICATE_TYPE, &x509_type, sizeof (x509_type) },
 	{ CKA_LABEL, "Cacert3 Here", 11 },
 	{ CKA_SUBJECT, (void *)test_cacert3_ca_subject, sizeof (test_cacert3_ca_subject) },
+	{ CKA_TRUSTED, &truev, sizeof (truev) },
 	{ CKA_ID, "ID1", 3 },
 	{ CKA_INVALID },
 };
 
+static CK_ATTRIBUTE cacert3_distrusted[] = {
+	{ CKA_VALUE, (void *)test_cacert3_ca_der, sizeof (test_cacert3_ca_der) },
+	{ CKA_CLASS, &certificate_class, sizeof (certificate_class) },
+	{ CKA_CERTIFICATE_TYPE, &x509_type, sizeof (x509_type) },
+	{ CKA_LABEL, "Another CaCert", 11 },
+	{ CKA_SUBJECT, (void *)test_cacert3_ca_subject, sizeof (test_cacert3_ca_subject) },
+	{ CKA_X_DISTRUSTED, &truev, sizeof (truev) },
+	{ CKA_INVALID },
+};
+
 static CK_ATTRIBUTE certificate_filter[] = {
 	{ CKA_CLASS, &certificate_class, sizeof (certificate_class) },
 	{ CKA_INVALID },
@@ -163,14 +218,15 @@ static CK_ATTRIBUTE extension_eku_invalid[] = {
 static void
 test_info_simple_certificate (CuTest *tc)
 {
-	CK_ATTRIBUTE *value;
+	void *value;
+	size_t length;
 	CK_RV rv;
 
 	setup (tc);
 
 	CuAssertPtrNotNull (tc, test.ex.asn1_defs);
 
-	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_authority_attrs);
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_trusted);
 	mock_module_add_object (MOCK_SLOT_ONE_ID, extension_eku_server_client);
 
 	p11_kit_iter_add_callback (test.iter, p11_extract_info_load_filter, &test.ex, NULL);
@@ -182,9 +238,9 @@ test_info_simple_certificate (CuTest *tc)
 
 	CuAssertIntEquals (tc, CKO_CERTIFICATE, test.ex.klass);
 	CuAssertPtrNotNull (tc, test.ex.attrs);
-	value = p11_attrs_find_valid (test.ex.attrs, CKA_VALUE);
+	value = p11_attrs_find_value (test.ex.attrs, CKA_VALUE, &length);
 	CuAssertPtrNotNull (tc, value);
-	CuAssertTrue (tc, memcmp (value->pValue, test_cacert3_ca_der, value->ulValueLen) == 0);
+	CuAssertTrue (tc, memcmp (value, test_cacert3_ca_der, length) == 0);
 	CuAssertPtrNotNull (tc, test.ex.cert_der);
 	CuAssertTrue (tc, memcmp (test.ex.cert_der, test_cacert3_ca_der, test.ex.cert_len) == 0);
 	CuAssertPtrNotNull (tc, test.ex.cert_asn);
@@ -202,7 +258,7 @@ test_info_limit_purposes (CuTest *tc)
 
 	setup (tc);
 
-	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_authority_attrs);
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_trusted);
 	mock_module_add_object (MOCK_SLOT_ONE_ID, extension_eku_server_client);
 
 	/* This should not match the above, with the stapled certificat ext */
@@ -227,7 +283,7 @@ test_info_invalid_purposes (CuTest *tc)
 
 	setup (tc);
 
-	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_authority_attrs);
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_trusted);
 	mock_module_add_object (MOCK_SLOT_ONE_ID, extension_eku_invalid);
 
 	p11_kit_iter_add_callback (test.iter, p11_extract_info_load_filter, &test.ex, NULL);
@@ -252,7 +308,7 @@ test_info_skip_non_certificate (CuTest *tc)
 
 	setup (tc);
 
-	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_authority_attrs);
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_trusted);
 
 	p11_kit_iter_add_callback (test.iter, p11_extract_info_load_filter, &test.ex, NULL);
 	p11_kit_iter_begin_with (test.iter, &test.module, 0, 0);
@@ -279,7 +335,7 @@ test_limit_to_purpose_match (CuTest *tc)
 
 	setup (tc);
 
-	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_authority_attrs);
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_trusted);
 	mock_module_add_object (MOCK_SLOT_ONE_ID, extension_eku_server_client);
 
 	p11_extract_info_limit_purpose (&test.ex, P11_OID_SERVER_AUTH_STR);
@@ -303,7 +359,7 @@ test_limit_to_purpose_no_match (CuTest *tc)
 
 	setup (tc);
 
-	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_authority_attrs);
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_trusted);
 	mock_module_add_object (MOCK_SLOT_ONE_ID, extension_eku_server_client);
 
 	p11_extract_info_limit_purpose (&test.ex, "3.3.3.3");
@@ -320,6 +376,146 @@ test_limit_to_purpose_no_match (CuTest *tc)
 	teardown (tc);
 }
 
+static void
+test_duplicate_extract (CuTest *tc)
+{
+	CK_ATTRIBUTE certificate = { CKA_CLASS, &certificate_class, sizeof (certificate_class) };
+	CK_RV rv;
+
+	setup (tc);
+
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_trusted);
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_distrusted);
+
+	p11_kit_iter_add_callback (test.iter, p11_extract_info_load_filter, &test.ex, NULL);
+	p11_kit_iter_add_filter (test.iter, &certificate, 1);
+	p11_kit_iter_begin_with (test.iter, &test.module, 0, 0);
+
+	rv = p11_kit_iter_next (test.iter);
+	CuAssertIntEquals (tc, CKR_OK, rv);
+
+	rv = p11_kit_iter_next (test.iter);
+	CuAssertIntEquals (tc, CKR_OK, rv);
+
+	rv = p11_kit_iter_next (test.iter);
+	CuAssertIntEquals (tc, CKR_CANCEL, rv);
+
+	teardown (tc);
+}
+
+static void
+test_duplicate_collapse (CuTest *tc)
+{
+	CK_ATTRIBUTE certificate = { CKA_CLASS, &certificate_class, sizeof (certificate_class) };
+	CK_RV rv;
+
+	setup (tc);
+
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_trusted);
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_distrusted);
+
+	test.ex.flags = P11_EXTRACT_COLLAPSE;
+	p11_kit_iter_add_callback (test.iter, p11_extract_info_load_filter, &test.ex, NULL);
+	p11_kit_iter_add_filter (test.iter, &certificate, 1);
+	p11_kit_iter_begin_with (test.iter, &test.module, 0, 0);
+
+	rv = p11_kit_iter_next (test.iter);
+	CuAssertIntEquals (tc, CKR_OK, rv);
+
+	rv = p11_kit_iter_next (test.iter);
+	CuAssertIntEquals (tc, CKR_CANCEL, rv);
+
+	teardown (tc);
+}
+
+static void
+test_trusted_match (CuTest *tc)
+{
+	CK_ATTRIBUTE certificate = { CKA_CLASS, &certificate_class, sizeof (certificate_class) };
+	CK_BBOOL boolv;
+	CK_RV rv;
+
+	setup (tc);
+
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_trusted);
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_distrusted);
+
+	test.ex.flags = P11_EXTRACT_ANCHORS;
+	p11_kit_iter_add_callback (test.iter, p11_extract_info_load_filter, &test.ex, NULL);
+	p11_kit_iter_add_filter (test.iter, &certificate, 1);
+	p11_kit_iter_begin_with (test.iter, &test.module, 0, 0);
+
+	rv = p11_kit_iter_next (test.iter);
+	CuAssertIntEquals (tc, CKR_OK, rv);
+
+	if (!p11_attrs_find_bool (test.ex.attrs, CKA_TRUSTED, &boolv))
+		boolv = CK_FALSE;
+	CuAssertIntEquals (tc, CK_TRUE, boolv);
+
+	rv = p11_kit_iter_next (test.iter);
+	CuAssertIntEquals (tc, CKR_CANCEL, rv);
+
+	teardown (tc);
+}
+
+static void
+test_distrust_match (CuTest *tc)
+{
+	CK_ATTRIBUTE certificate = { CKA_CLASS, &certificate_class, sizeof (certificate_class) };
+	CK_BBOOL boolv;
+	CK_RV rv;
+
+	setup (tc);
+
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_trusted);
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_distrusted);
+
+	test.ex.flags = P11_EXTRACT_BLACKLIST;
+	p11_kit_iter_add_callback (test.iter, p11_extract_info_load_filter, &test.ex, NULL);
+	p11_kit_iter_add_filter (test.iter, &certificate, 1);
+	p11_kit_iter_begin_with (test.iter, &test.module, 0, 0);
+
+	rv = p11_kit_iter_next (test.iter);
+	CuAssertIntEquals (tc, CKR_OK, rv);
+
+	if (!p11_attrs_find_bool (test.ex.attrs, CKA_X_DISTRUSTED, &boolv))
+		boolv = CK_FALSE;
+	CuAssertIntEquals (tc, CK_TRUE, boolv);
+
+	rv = p11_kit_iter_next (test.iter);
+	CuAssertIntEquals (tc, CKR_CANCEL, rv);
+
+	teardown (tc);
+}
+
+static void
+test_anytrust_match (CuTest *tc)
+{
+	CK_ATTRIBUTE certificate = { CKA_CLASS, &certificate_class, sizeof (certificate_class) };
+	CK_RV rv;
+
+	setup (tc);
+
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_trusted);
+	mock_module_add_object (MOCK_SLOT_ONE_ID, cacert3_distrusted);
+
+	test.ex.flags =  P11_EXTRACT_ANCHORS | P11_EXTRACT_BLACKLIST;
+	p11_kit_iter_add_callback (test.iter, p11_extract_info_load_filter, &test.ex, NULL);
+	p11_kit_iter_add_filter (test.iter, &certificate, 1);
+	p11_kit_iter_begin_with (test.iter, &test.module, 0, 0);
+
+	rv = p11_kit_iter_next (test.iter);
+	CuAssertIntEquals (tc, CKR_OK, rv);
+
+	rv = p11_kit_iter_next (test.iter);
+	CuAssertIntEquals (tc, CKR_OK, rv);
+
+	rv = p11_kit_iter_next (test.iter);
+	CuAssertIntEquals (tc, CKR_CANCEL, rv);
+
+	teardown (tc);
+}
+
 int
 main (void)
 {
@@ -334,12 +530,19 @@ main (void)
 
 	SUITE_ADD_TEST (suite, test_file_name_for_label);
 	SUITE_ADD_TEST (suite, test_file_name_for_class);
+	SUITE_ADD_TEST (suite, test_comment_for_label);
+	SUITE_ADD_TEST (suite, test_comment_not_enabled);
 	SUITE_ADD_TEST (suite, test_info_simple_certificate);
 	SUITE_ADD_TEST (suite, test_info_limit_purposes);
 	SUITE_ADD_TEST (suite, test_info_invalid_purposes);
 	SUITE_ADD_TEST (suite, test_info_skip_non_certificate);
 	SUITE_ADD_TEST (suite, test_limit_to_purpose_match);
 	SUITE_ADD_TEST (suite, test_limit_to_purpose_no_match);
+	SUITE_ADD_TEST (suite, test_duplicate_extract);
+	SUITE_ADD_TEST (suite, test_duplicate_collapse);
+	SUITE_ADD_TEST (suite, test_trusted_match);
+	SUITE_ADD_TEST (suite, test_distrust_match);
+	SUITE_ADD_TEST (suite, test_anytrust_match);
 
 	CuSuiteRun (suite);
 	CuSuiteSummary (suite, output);
