@@ -91,6 +91,8 @@ write_zero_file (CuTest *tc,
 	CuAssertTrue (tc, fd != -1);
 	res = close (fd);
 	CuAssertTrue (tc, res >= 0);
+
+	free (filename);
 }
 
 static void
@@ -190,6 +192,54 @@ test_file_overwrite (CuTest *tc)
 	teardown (tc);
 }
 
+static void
+test_file_auto_empty (CuTest *tc)
+{
+	p11_save_file *file;
+	char *filename;
+	bool ret;
+
+	setup (tc);
+
+	if (asprintf (&filename, "%s/%s", test.directory, "extract-file") < 0)
+		CuFail (tc, "asprintf() failed");
+
+	file = p11_save_open_file (filename, 0);
+	CuAssertPtrNotNull (tc, file);
+
+	ret = p11_save_write_and_finish (file, NULL, -1);
+	CuAssertIntEquals (tc, true, ret);
+	free (filename);
+
+	test_check_file (tc, test.directory, "extract-file", SRCDIR "/files/empty-file");
+
+	teardown (tc);
+}
+
+static void
+test_file_auto_length (CuTest *tc)
+{
+	p11_save_file *file;
+	char *filename;
+	bool ret;
+
+	setup (tc);
+
+	if (asprintf (&filename, "%s/%s", test.directory, "extract-file") < 0)
+		CuFail (tc, "asprintf() failed");
+
+	file = p11_save_open_file (filename, 0);
+	CuAssertPtrNotNull (tc, file);
+
+	ret = p11_save_write_and_finish (file, "The simple string is hairy", -1);
+	CuAssertIntEquals (tc, true, ret);
+	free (filename);
+
+	test_check_file (tc, test.directory, "extract-file", SRCDIR "/files/simple-string");
+
+	teardown (tc);
+}
+
 static void
 test_write_with_null (CuTest *tc)
 {
@@ -256,7 +306,7 @@ test_directory_empty (CuTest *tc)
 
 	test_check_directory (tc, subdir, (NULL, NULL));
 
-	rmdir (subdir);
+	CuAssertTrue (tc, rmdir (subdir) >= 0);
 	free (subdir);
 
 	teardown (tc);
@@ -307,7 +357,7 @@ test_directory_files (CuTest *tc)
 	test_check_symlink (tc, subdir, "link.ext", "/the/destination");
 #endif
 
-	rmdir (subdir);
+	CuAssertTrue (tc, rmdir (subdir) >= 0);
 	free (subdir);
 
 	teardown (tc);
@@ -389,7 +439,7 @@ test_directory_dups (CuTest *tc)
 	test_check_symlink (tc, subdir, "link.1", "/destination2");
 #endif
 
-	rmdir (subdir);
+	CuAssertTrue (tc, rmdir (subdir) >= 0);
 	free (subdir);
 
 	teardown (tc);
@@ -439,18 +489,15 @@ test_directory_overwrite (CuTest *tc)
 	if (asprintf (&subdir, "%s/%s", test.directory, "extract-dir") < 0)
 		CuFail (tc, "asprintf() failed");
 
-#ifdef OS_UNIX
-	if (mkdir (subdir, S_IRWXU) < 0)
-#else
-	if (mkdir (subdir) < 0)
-#endif
-		CuFail (tc, "mkdir() failed");
-
 	/* Some initial files into this directory, which get overwritten */
-	write_zero_file (tc, subdir, "file.txt");
-	write_zero_file (tc, subdir, "another-file");
-	write_zero_file (tc, subdir, "third-file");
+	dir = p11_save_open_directory (subdir, 0);
+	ret = p11_save_write_and_finish (p11_save_open_file_in (dir, "file", ".txt", NULL), "", 0) &&
+	      p11_save_write_and_finish (p11_save_open_file_in (dir, "another-file", NULL, NULL), "", 0) &&
+	      p11_save_write_and_finish (p11_save_open_file_in (dir, "third-file", NULL, NULL), "", 0) &&
+	      p11_save_finish_directory (dir, true);
+	CuAssertTrue (tc, ret && dir);
 
+	/* Now the actual test, using the same directory */
 	dir = p11_save_open_directory (subdir, P11_SAVE_OVERWRITE);
 	CuAssertPtrNotNull (tc, dir);
 
@@ -477,7 +524,7 @@ test_directory_overwrite (CuTest *tc)
 	test_check_data (tc, subdir, "file.txt", test_text, strlen (test_text));
 	test_check_data (tc, subdir, "file.1.txt", test_text, 10);
 
-	rmdir (subdir);
+	CuAssertTrue (tc, rmdir (subdir) >= 0);
 	free (subdir);
 
 	teardown (tc);
@@ -496,6 +543,8 @@ main (void)
 	SUITE_ADD_TEST (suite, test_file_exists);
 	SUITE_ADD_TEST (suite, test_file_bad_directory);
 	SUITE_ADD_TEST (suite, test_file_overwrite);
+	SUITE_ADD_TEST (suite, test_file_auto_empty);
+	SUITE_ADD_TEST (suite, test_file_auto_length);
 	SUITE_ADD_TEST (suite, test_write_with_null);
 	SUITE_ADD_TEST (suite, test_write_and_finish_with_null);
 	SUITE_ADD_TEST (suite, test_file_abort);
