@@ -36,13 +36,16 @@
 
 #define CRYPTOKI_EXPORTS
 
+#include "array.h"
 #include "attrs.h"
 #define P11_DEBUG_FLAG P11_DEBUG_TRUST
 #include "debug.h"
 #include "dict.h"
 #include "library.h"
 #include "module.h"
+#include "parser.h"
 #include "pkcs11.h"
+#include "pkcs11x.h"
 #include "session.h"
 #include "token.h"
 
@@ -53,19 +56,16 @@
 
 #define MANUFACTURER_ID         "PKCS#11 Kit                     "
 #define LIBRARY_DESCRIPTION     "PKCS#11 Kit Trust Module        "
-#define SLOT_DESCRIPTION        "System Certificates, Trust Anchors, and Black Lists             "
-#define TOKEN_LABEL             "System Trust Anchors and Policy "
-#define TOKEN_MODEL             "PKCS#11 Kit      "
+#define TOKEN_MODEL             "p11-kit-trust    "
 #define TOKEN_SERIAL_NUMBER     "1                "
 
-/* Arbitrary non-zero and non-one choice */
-#define SYSTEM_SLOT_ID   18UL
+/* Initial slot id: non-zero and non-one */
+#define BASE_SLOT_ID   18UL
 
 static struct _Shared {
 	p11_dict *sessions;
-	p11_token *token;
-	char *anchor_paths;
-	char *certificate_paths;
+	p11_array *tokens;
+	char *paths;
 } gl = { NULL, NULL };
 
 /* Used during FindObjects */
@@ -104,6 +104,147 @@ lookup_session (CK_SESSION_HANDLE handle,
 	return CKR_OK;
 }
 
+static CK_ATTRIBUTE *
+lookup_object_inlock (p11_session *session,
+                      CK_OBJECT_HANDLE handle,
+                      p11_index **index)
+{
+	CK_ATTRIBUTE *attrs;
+
+	assert (session != NULL);
+
+	attrs = p11_index_lookup (session->index, handle);
+	if (attrs) {
+		if (index)
+			*index = session->index;
+		return attrs;
+	}
+
+	attrs = p11_index_lookup (p11_token_index (session->token), handle);
+	if (attrs) {
+		if (index)
+			*index = p11_token_index (session->token);
+		return attrs;
+	}
+
+	return NULL;
+}
+
+
+static CK_RV
+lookup_slot_inlock (CK_SLOT_ID id,
+                    p11_token **token)
+{
+	/*
+	 * These are invalid inputs, that well behaved callers should
+	 * not produce, so have them fail precondations
+	 */
+
+	return_val_if_fail (gl.tokens != NULL,
+	                    CKR_CRYPTOKI_NOT_INITIALIZED);
+
+	return_val_if_fail (id >= BASE_SLOT_ID && id - BASE_SLOT_ID < gl.tokens->num,
+	                    CKR_SLOT_ID_INVALID);
+
+	if (token)
+		*token = gl.tokens->elem[id - BASE_SLOT_ID];
+	return CKR_OK;
+}
+
+static bool
+check_slot (CK_SLOT_ID id)
+{
+	bool ret;
+
+	p11_lock ();
+	ret = lookup_slot_inlock (id, NULL) == CKR_OK;
+	p11_unlock ();
+
+	return ret;
+}
+
+static bool
+create_tokens_inlock (p11_array *tokens,
+                      const char *paths)
+{
+	/*
+	 * TRANSLATORS: These label strings are used in PKCS#11 URIs and
+	 * unfortunately cannot be marked translatable. If localization is
+	 * desired they should be translated in GUI applications. These
+	 * strings will not change arbitrarily.
+	 */
+
+	struct {
+		const char *prefix;
+		const char *label;
+	} labels[] = {
+		{ DATADIR, "Default Trust" },
+		{ SYSCONFDIR, "System Trust" },
+		{ NULL },
+	};
+
+	p11_token *token;
+	p11_token *check;
+	CK_SLOT_ID slot;
+	const char *path;
+	const char *label;
+	char *alloc;
+	char *remaining;
+	char *base;
+	char *pos;
+	int i;
+
+	p11_debug ("using paths: %s", paths);
+
+	alloc = remaining = strdup (paths);
+	return_val_if_fail (remaining != NULL, false);
+
+	while (remaining) {
+		path = remaining;
+		pos = strchr (remaining, ':');
+		if (pos == NULL) {
+			remaining = NULL;
+		} else {
+			pos[0] = '\0';
+			remaining = pos + 1;
+		}
+
+		if (path[0] != '\0') {
+			/* The slot for the new token */
+			slot = BASE_SLOT_ID + tokens->num;
+
+			label = NULL;
+			base = NULL;
+
+			/* Claim the various labels based on prefix */
+			for (i = 0; label == NULL && labels[i].prefix != NULL; i++) {
+				if (strncmp (path, labels[i].prefix, strlen (labels[i].prefix)) == 0) {
+					label = labels[i].label;
+					labels[i].label = NULL;
+				}
+			}
+
+			/* Didn't find a label above, then make one based on the path */
+			if (!label) {
+				label = base = p11_basename (path);
+				return_val_if_fail (base != NULL, false);
+			}
+
+			token = p11_token_new (slot, path, label);
+			return_val_if_fail (token != NULL, false);
+
+			if (!p11_array_push (tokens, token))
+				return_val_if_reached (false);
+
+			free (base);
+			assert (lookup_slot_inlock (slot, &check) == CKR_OK && check == token);
+		}
+	}
+
+	free (alloc);
+	return true;
+}
+
 static void
 parse_argument (char *arg)
 {
@@ -115,13 +256,9 @@ parse_argument (char *arg)
 	else
 		*(value++) = 0;
 
-	if (strcmp (arg, "anchors") == 0) {
-		free (gl.anchor_paths);
-		gl.anchor_paths = value ? strdup (value) : NULL;
-
-	} else if (strcmp (arg, "certificates") == 0) {
-		free (gl.certificate_paths);
-		gl.certificate_paths = value ? strdup (value) : NULL;
+	if (strcmp (arg, "paths") == 0) {
+		free (gl.paths);
+		gl.paths = value ? strdup (value) : NULL;
 
 	} else {
 		p11_message ("unrecognized module argument: %s", arg);
@@ -219,15 +356,14 @@ sys_C_Finalize (CK_VOID_PTR reserved)
 				rv = CKR_CRYPTOKI_NOT_INITIALIZED;
 
 			} else {
-				free (gl.certificate_paths);
-				free (gl.anchor_paths);
-				gl.certificate_paths = gl.anchor_paths = NULL;
+				free (gl.paths);
+				gl.paths = NULL;
 
 				p11_dict_free (gl.sessions);
 				gl.sessions = NULL;
 
-				p11_token_free (gl.token);
-				gl.token = NULL;
+				p11_array_free (gl.tokens);
+				gl.tokens = NULL;
 
 				rv = CKR_OK;
 			}
@@ -290,10 +426,11 @@ sys_C_Initialize (CK_VOID_PTR init_args)
 			                            p11_dict_ulongptr_equal,
 			                            NULL, p11_session_free);
 
-			gl.token = p11_token_new (gl.anchor_paths ? gl.anchor_paths : SYSTEM_ANCHORS,
-			                          gl.certificate_paths ? gl.certificate_paths : SYSTEM_CERTIFICATES);
+			gl.tokens = p11_array_new ((p11_destroyer)p11_token_free);
+			if (gl.tokens && !create_tokens_inlock (gl.tokens, gl.paths ? gl.paths : TRUST_PATHS))
+				gl.tokens = NULL;
 
-			if (gl.sessions == NULL || gl.token == NULL) {
+			if (gl.sessions == NULL || gl.tokens == NULL) {
 				warn_if_reached ();
 				rv = CKR_GENERAL_ERROR;
 			}
@@ -358,6 +495,7 @@ sys_C_GetSlotList (CK_BBOOL token_present,
                    CK_ULONG_PTR count)
 {
 	CK_RV rv = CKR_OK;
+	int i;
 
 	return_val_if_fail (count != NULL, CKR_ARGUMENTS_BAD);
 
@@ -374,16 +512,17 @@ sys_C_GetSlotList (CK_BBOOL token_present,
 		/* already failed */
 
 	} else if (!slot_list) {
-		*count = 1;
+		*count = gl.tokens->num;
 		rv = CKR_OK;
 
-	} else if (*count < 1) {
-		*count = 1;
+	} else if (*count < gl.tokens->num) {
+		*count = gl.tokens->num;
 		rv = CKR_BUFFER_TOO_SMALL;
 
 	} else {
-		slot_list[0] = SYSTEM_SLOT_ID;
-		*count = 1;
+		for (i = 0; i < gl.tokens->num; i++)
+			slot_list[i] = BASE_SLOT_ID + i;
+		*count = gl.tokens->num;
 		rv = CKR_OK;
 	}
 
@@ -397,30 +536,35 @@ sys_C_GetSlotInfo (CK_SLOT_ID id,
                    CK_SLOT_INFO_PTR info)
 {
 	CK_RV rv = CKR_OK;
+	p11_token *token;
+	const char *path;
+	size_t length;
 
-	return_val_if_fail (id == SYSTEM_SLOT_ID, CKR_SLOT_ID_INVALID);
 	return_val_if_fail (info != NULL, CKR_ARGUMENTS_BAD);
 
 	p11_debug ("in");
-
 	p11_lock ();
 
-		if (!gl.sessions)
-			rv = CKR_CRYPTOKI_NOT_INITIALIZED;
-
-	p11_unlock ();
-
+	rv = lookup_slot_inlock (id, &token);
 	if (rv == CKR_OK) {
 		memset (info, 0, sizeof (*info));
 		info->firmwareVersion.major = 0;
 		info->firmwareVersion.minor = 0;
-		info->hardwareVersion.major = 0;
-		info->hardwareVersion.minor = 0;
+		info->hardwareVersion.major = PACKAGE_MAJOR;
+		info->hardwareVersion.minor = PACKAGE_MINOR;
 		info->flags = CKF_TOKEN_PRESENT;
 		strncpy ((char*)info->manufacturerID, MANUFACTURER_ID, 32);
-		strncpy ((char*)info->slotDescription, SLOT_DESCRIPTION, 64);
+
+		/* If too long, copy the first 64 characters into buffer */
+		path = p11_token_get_path (token);
+		length = strlen (path);
+		if (length > sizeof (info->slotDescription))
+			length = sizeof (info->slotDescription);
+		memset (info->slotDescription, ' ', sizeof (info->slotDescription));
+		memcpy (info->slotDescription, path, length);
 	}
 
+	p11_unlock ();
 	p11_debug ("out: 0x%lx", rv);
 
 	return rv;
@@ -431,28 +575,25 @@ sys_C_GetTokenInfo (CK_SLOT_ID id,
                     CK_TOKEN_INFO_PTR info)
 {
 	CK_RV rv = CKR_OK;
+	p11_token *token;
+	const char *label;
+	size_t length;
 
-	return_val_if_fail (id == SYSTEM_SLOT_ID, CKR_SLOT_ID_INVALID);
 	return_val_if_fail (info != NULL, CKR_ARGUMENTS_BAD);
 
 	p11_debug ("in");
 
 	p11_lock ();
 
-		if (!gl.sessions)
-			rv = CKR_CRYPTOKI_NOT_INITIALIZED;
-
-	p11_unlock ();
-
+	rv = lookup_slot_inlock (id, &token);
 	if (rv == CKR_OK) {
 		memset (info, 0, sizeof (*info));
 		info->firmwareVersion.major = 0;
 		info->firmwareVersion.minor = 0;
-		info->hardwareVersion.major = 0;
-		info->hardwareVersion.minor = 0;
+		info->hardwareVersion.major = PACKAGE_MAJOR;
+		info->hardwareVersion.minor = PACKAGE_MINOR;
 		info->flags = CKF_TOKEN_INITIALIZED | CKF_WRITE_PROTECTED;
 		strncpy ((char*)info->manufacturerID, MANUFACTURER_ID, 32);
-		strncpy ((char*)info->label, TOKEN_LABEL, 32);
 		strncpy ((char*)info->model, TOKEN_MODEL, 16);
 		strncpy ((char*)info->serialNumber, TOKEN_SERIAL_NUMBER, 16);
 		info->ulMaxSessionCount = CK_EFFECTIVELY_INFINITE;
@@ -465,8 +606,17 @@ sys_C_GetTokenInfo (CK_SLOT_ID id,
 		info->ulFreePublicMemory = CK_UNAVAILABLE_INFORMATION;
 		info->ulTotalPrivateMemory = CK_UNAVAILABLE_INFORMATION;
 		info->ulFreePrivateMemory = CK_UNAVAILABLE_INFORMATION;
+
+		/* If too long, copy the first 32 characters into buffer */
+		label = p11_token_get_label (token);
+		length = strlen (label);
+		if (length > sizeof (info->label))
+			length = sizeof (info->label);
+		memset (info->label, ' ', sizeof (info->label));
+		memcpy (info->label, label, length);
 	}
 
+	p11_unlock ();
 	p11_debug ("out: 0x%lx", rv);
 
 	return rv;
@@ -494,8 +644,8 @@ sys_C_GetMechanismInfo (CK_SLOT_ID id,
                         CK_MECHANISM_TYPE type,
                         CK_MECHANISM_INFO_PTR info)
 {
-	return_val_if_fail (id == SYSTEM_SLOT_ID, CKR_SLOT_ID_INVALID);
 	return_val_if_fail (info != NULL, CKR_ARGUMENTS_BAD);
+	return_val_if_fail (check_slot (id), CKR_SLOT_ID_INVALID);
 	return_val_if_reached (CKR_MECHANISM_INVALID);
 }
 
@@ -505,7 +655,7 @@ sys_C_InitToken (CK_SLOT_ID id,
                  CK_ULONG pin_len,
                  CK_UTF8CHAR_PTR label)
 {
-	return_val_if_fail (id == SYSTEM_SLOT_ID, CKR_SLOT_ID_INVALID);
+	return_val_if_fail (check_slot (id), CKR_SLOT_ID_INVALID);
 	return_val_if_reached (CKR_TOKEN_WRITE_PROTECTED);
 }
 
@@ -526,17 +676,19 @@ sys_C_OpenSession (CK_SLOT_ID id,
                    CK_SESSION_HANDLE_PTR handle)
 {
 	p11_session *session;
+	p11_token *token;
 	CK_RV rv = CKR_OK;
 
-	return_val_if_fail (id == SYSTEM_SLOT_ID, CKR_SLOT_ID_INVALID);
+	return_val_if_fail (check_slot (id), CKR_SLOT_ID_INVALID);
 	return_val_if_fail (handle != NULL, CKR_ARGUMENTS_BAD);
 
 	p11_debug ("in");
 
 	p11_lock ();
 
-		if (!gl.sessions) {
-			rv = CKR_CRYPTOKI_NOT_INITIALIZED;
+		rv = lookup_slot_inlock (id, &token);
+		if (rv != CKR_OK) {
+			/* fail below */;
 
 		} else if (!(flags & CKF_SERIAL_SESSION)) {
 			rv = CKR_SESSION_PARALLEL_NOT_SUPPORTED;
@@ -545,7 +697,7 @@ sys_C_OpenSession (CK_SLOT_ID id,
 			rv = CKR_TOKEN_WRITE_PROTECTED;
 
 		} else {
-			session = p11_session_new (gl.token);
+			session = p11_session_new (token);
 			if (p11_dict_set (gl.sessions, &session->handle, session)) {
 				rv = CKR_OK;
 				*handle = session->handle;
@@ -592,20 +744,23 @@ sys_C_CloseSession (CK_SESSION_HANDLE handle)
 static CK_RV
 sys_C_CloseAllSessions (CK_SLOT_ID id)
 {
-	CK_RV rv = CKR_OK;
-
-	return_val_if_fail (id == SYSTEM_SLOT_ID, CKR_SLOT_ID_INVALID);
+	CK_SESSION_HANDLE *handle;
+	p11_session *session;
+	p11_token *token;
+	p11_dictiter iter;
+	CK_RV rv;
 
 	p11_debug ("in");
 
 	p11_lock ();
 
-		if (!gl.sessions) {
-			rv = CKR_CRYPTOKI_NOT_INITIALIZED;
-
-		} else {
-			p11_dict_clear (gl.sessions);
-			rv = CKR_OK;
+		rv = lookup_slot_inlock (id, &token);
+		if (rv == CKR_OK) {
+			p11_dict_iterate (gl.sessions, &iter);
+			while (p11_dict_next (&iter, (void **)&handle, (void **)&session)) {
+				if (session->token == token)
+					p11_dict_remove (gl.sessions, handle);
+			}
 		}
 
 	p11_unlock ();
@@ -631,6 +786,7 @@ static CK_RV
 sys_C_GetSessionInfo (CK_SESSION_HANDLE handle,
                       CK_SESSION_INFO_PTR info)
 {
+	p11_session *session;
 	CK_RV rv;
 
 	return_val_if_fail (info != NULL, CKR_ARGUMENTS_BAD);
@@ -639,17 +795,17 @@ sys_C_GetSessionInfo (CK_SESSION_HANDLE handle,
 
 	p11_lock ();
 
-		rv = lookup_session (handle, NULL);
+		rv = lookup_session (handle, &session);
+		if (rv == CKR_OK) {
+			info->flags = CKF_SERIAL_SESSION;
+			info->state = CKS_RO_PUBLIC_SESSION;
+			info->slotID = p11_token_get_slot (session->token);
+			info->ulDeviceError = 0;
+		}
+
 
 	p11_unlock ();
 
-	if (rv == CKR_OK) {
-		info->flags = CKF_SERIAL_SESSION;
-		info->slotID = SYSTEM_SLOT_ID;
-		info->state = CKS_RO_PUBLIC_SESSION;
-		info->ulDeviceError = 0;
-	}
-
 	p11_debug ("out: 0x%lx", rv);
 
 	return rv;
@@ -714,7 +870,6 @@ sys_C_CreateObject (CK_SESSION_HANDLE handle,
                     CK_ULONG count,
                     CK_OBJECT_HANDLE_PTR new_object)
 {
-	CK_ATTRIBUTE *attrs;
 	p11_session *session;
 	CK_BBOOL token;
 	CK_RV rv;
@@ -731,10 +886,8 @@ sys_C_CreateObject (CK_SESSION_HANDLE handle,
 				rv = CKR_TOKEN_WRITE_PROTECTED;
 		}
 
-		if (rv == CKR_OK) {
-			attrs = p11_attrs_buildn (NULL, template, count);
-			rv = p11_session_add_object (session, attrs, new_object);
-		}
+		if (rv == CKR_OK)
+			rv = p11_index_add (session->index, template, count, new_object);
 
 	p11_unlock ();
 
@@ -766,7 +919,7 @@ sys_C_CopyObject (CK_SESSION_HANDLE handle,
 
 		rv = lookup_session (handle, &session);
 		if (rv == CKR_OK) {
-			original = p11_session_get_object (session, object, NULL);
+			original = lookup_object_inlock (session, object, NULL);
 			if (original == NULL)
 				rv = CKR_OBJECT_HANDLE_INVALID;
 		}
@@ -780,7 +933,7 @@ sys_C_CopyObject (CK_SESSION_HANDLE handle,
 			attrs = p11_attrs_dup (original);
 			attrs = p11_attrs_buildn (attrs, template, count);
 			attrs = p11_attrs_build (attrs, &token, NULL);
-			rv = p11_session_add_object (session, attrs, new_object);
+			rv = p11_index_take (session->index, attrs, new_object);
 		}
 
 	p11_unlock ();
@@ -802,8 +955,13 @@ sys_C_DestroyObject (CK_SESSION_HANDLE handle,
 	p11_lock ();
 
 		rv = lookup_session (handle, &session);
-		if (rv == CKR_OK)
-			rv = p11_session_del_object (session, object);
+		if (rv == CKR_OK) {
+			rv = p11_index_remove (session->index, object);
+			if (rv == CKR_OBJECT_HANDLE_INVALID) {
+				if (p11_index_lookup (p11_token_index (session->token), object))
+					rv = CKR_TOKEN_WRITE_PROTECTED;
+			}
+		}
 
 	p11_unlock ();
 
@@ -828,7 +986,7 @@ sys_C_GetObjectSize (CK_SESSION_HANDLE handle,
 
 		rv = lookup_session (handle, &session);
 		if (rv == CKR_OK) {
-			if (p11_session_get_object (session, object, NULL)) {
+			if (lookup_object_inlock (session, object, NULL)) {
 				*size = CK_UNAVAILABLE_INFORMATION;
 				rv = CKR_OK;
 			} else {
@@ -853,16 +1011,17 @@ sys_C_GetAttributeValue (CK_SESSION_HANDLE handle,
 	CK_ATTRIBUTE *result;
 	CK_ATTRIBUTE *attr;
 	p11_session *session;
+	char *string;
 	CK_ULONG i;
 	CK_RV rv;
 
-	p11_debug ("in");
+	p11_debug ("in: %lu, %lu", handle, object);
 
 	p11_lock ();
 
 		rv = lookup_session (handle, &session);
 		if (rv == CKR_OK) {
-			attrs = p11_session_get_object (session, object, NULL);
+			attrs = lookup_object_inlock (session, object, NULL);
 			if (attrs == NULL)
 				rv = CKR_OBJECT_HANDLE_INVALID;
 		}
@@ -895,7 +1054,11 @@ sys_C_GetAttributeValue (CK_SESSION_HANDLE handle,
 
 	p11_unlock ();
 
-	p11_debug ("out: 0x%lx", rv);
+	if (p11_debugging) {
+		string = p11_attrs_to_string (template, count);
+		p11_debug ("out: 0x%lx %s", rv, string);
+		free (string);
+	}
 
 	return rv;
 }
@@ -914,8 +1077,13 @@ sys_C_SetAttributeValue (CK_SESSION_HANDLE handle,
 	p11_lock ();
 
 		rv = lookup_session (handle, &session);
-		if (rv == CKR_OK)
-			rv = p11_session_set_object (session, object, template, count);
+		if (rv == CKR_OK) {
+			rv = p11_index_set (session->index, object, template, count);
+			if (rv == CKR_OBJECT_HANDLE_INVALID) {
+				if (p11_index_lookup (p11_token_index (session->token), object))
+					rv = CKR_TOKEN_WRITE_PROTECTED;
+			}
+		}
 
 	p11_unlock ();
 
@@ -924,23 +1092,69 @@ sys_C_SetAttributeValue (CK_SESSION_HANDLE handle,
 	return rv;
 }
 
+static CK_ATTRIBUTE *
+work_around_broken_nss_serial_number_lookups (CK_ATTRIBUTE *attrs)
+{
+	/*
+	 * WORKAROUND: NSS calls us asking for CKA_SERIAL_NUMBER items that are
+	 * not DER encoded. It shouldn't be doing this. We never return any certificate
+	 * serial numbers that are not DER encoded.
+	 *
+	 * So work around the issue here while the NSS guys fix this issue.
+	 * This code should be removed in future versions.
+	 */
+
+	CK_OBJECT_CLASS klass;
+	CK_ATTRIBUTE *serial;
+	unsigned char *der;
+	size_t der_len;
+	int len_len;
+
+	if (!p11_attrs_find_ulong (attrs, CKA_CLASS, &klass) ||
+	    klass != CKO_NSS_TRUST)
+		return attrs;
+
+	serial = p11_attrs_find_valid (attrs, CKA_SERIAL_NUMBER);
+	if (!serial || p11_asn1_tlv_length (serial->pValue, serial->ulValueLen) >= 0)
+		return attrs;
+
+	p11_debug ("working around serial number lookup that's not DER encoded");
+
+	/* Assumption that 32 bytes is more than enough to store a ulong */
+	der_len = 1 + 32 + serial->ulValueLen;
+	der = malloc (der_len);
+	return_val_if_fail (der != NULL, NULL);
+
+	der[0] = ASN1_TAG_INTEGER | ASN1_CLASS_UNIVERSAL;
+	len_len = der_len - 1;
+	asn1_length_der (serial->ulValueLen, der + 1, &len_len);
+	assert (len_len < der_len - serial->ulValueLen);
+	memcpy (der + 1 + len_len, serial->pValue, serial->ulValueLen);
+	der_len = 1 + len_len + serial->ulValueLen;
+
+	return p11_attrs_take (attrs, CKA_SERIAL_NUMBER, der, der_len);
+}
+
 static CK_RV
 sys_C_FindObjectsInit (CK_SESSION_HANDLE handle,
                        CK_ATTRIBUTE_PTR template,
                        CK_ULONG count)
 {
-	CK_OBJECT_HANDLE *handle_ptr;
+	p11_index *indices[2] = { NULL, NULL };
 	CK_BBOOL want_token_objects;
 	CK_BBOOL want_session_objects;
 	CK_BBOOL token;
-	p11_dict *objects;
 	FindObjects *find;
 	p11_session *session;
-	p11_dictiter iter;
-	CK_ULONG i;
+	char *string;
 	CK_RV rv;
+	int n = 0;
 
-	p11_debug ("in");
+	if (p11_debugging) {
+		string = p11_attrs_to_string (template, count);
+		p11_debug ("in: %lu, %s", handle, string);
+		free (string);
+	}
 
 	p11_lock ();
 
@@ -956,52 +1170,33 @@ sys_C_FindObjectsInit (CK_SESSION_HANDLE handle,
 		rv = lookup_session (handle, &session);
 
 		/* Refresh from disk if this session hasn't yet */
-		if (rv == CKR_OK && want_token_objects && !session->loaded) {
-			session->loaded = CK_TRUE;
-			p11_token_load (gl.token);
-		}
-
 		if (rv == CKR_OK) {
-			objects = p11_token_objects (gl.token);
+			if (want_session_objects)
+				indices[n++] = session->index;
+			if (want_token_objects) {
+				if (!session->loaded)
+					p11_token_load (session->token);
+				session->loaded = CK_TRUE;
+				indices[n++] = p11_token_index (session->token);
+			}
 
 			find = calloc (1, sizeof (FindObjects));
 			warn_if_fail (find != NULL);
 
-			/* Make a copy of what we're matching */
+			/* Make a snapshot of what we're matching */
 			if (find) {
 				find->match = p11_attrs_buildn (NULL, template, count);
 				warn_if_fail (find->match != NULL);
 
+				find->match = work_around_broken_nss_serial_number_lookups (find->match);
+				warn_if_fail (find->match != NULL);
+
 				/* Build a session snapshot of all objects */
 				find->iterator = 0;
-				count = p11_dict_size (objects) + p11_dict_size (session->objects) + 1;
-				find->snapshot = calloc (count, sizeof (CK_OBJECT_HANDLE));
+				find->snapshot = p11_index_snapshot (indices[0], indices[1], template, count);
 				warn_if_fail (find->snapshot != NULL);
 			}
 
-			if (find && find->snapshot) {
-				i = 0;
-
-				if (want_token_objects) {
-					p11_dict_iterate (objects, &iter);
-					for ( ; p11_dict_next (&iter, (void *)&handle_ptr, NULL); i++) {
-						assert (i < count);
-						find->snapshot[i] = *handle_ptr;
-					}
-				}
-
-				if (want_session_objects) {
-					p11_dict_iterate (session->objects, &iter);
-					for ( ; p11_dict_next (&iter, (void *)&handle_ptr, NULL); i++) {
-						assert (i < count);
-						find->snapshot[i] = *handle_ptr;
-					}
-				}
-
-				assert (i < count);
-				assert (find->snapshot[i] == 0UL);
-			}
-
 			if (!find || !find->snapshot || !find->match)
 				rv = CKR_HOST_MEMORY;
 			else
@@ -1026,11 +1221,12 @@ sys_C_FindObjects (CK_SESSION_HANDLE handle,
 	FindObjects *find = NULL;
 	p11_session *session;
 	CK_ULONG matched;
+	p11_index *index;
 	CK_RV rv;
 
 	return_val_if_fail (count != NULL, CKR_ARGUMENTS_BAD);
 
-	p11_debug ("in");
+	p11_debug ("in: %lu, %lu", handle, max_count);
 
 	p11_lock ();
 
@@ -1050,7 +1246,7 @@ sys_C_FindObjects (CK_SESSION_HANDLE handle,
 
 				find->iterator++;
 
-				attrs = p11_session_get_object (session, object, NULL);
+				attrs = lookup_object_inlock (session, object, &index);
 				if (attrs == NULL)
 					continue;
 
@@ -1065,7 +1261,7 @@ sys_C_FindObjects (CK_SESSION_HANDLE handle,
 
 	p11_unlock ();
 
-	p11_debug ("out: 0x%lx", rv);
+	p11_debug ("out: 0x%lx, %lu", handle, *count);
 
 	return rv;
 }
